@@ -2,6 +2,7 @@ import logging
 import multiprocessing
 import os
 import random
+import shutil
 import sys
 from pathlib import Path
 from typing import Dict, Optional
@@ -31,6 +32,11 @@ class Config:
     :py:func:`StepWithArtifacts.gather_artifacts`.
     """
 
+    log_path: Path
+    """
+    The path to the directory in which logs are stored.
+    """
+
     colors: bool
     """
     Whether to use colored output or not for the output.
@@ -117,6 +123,7 @@ class Config:
     def __init__(
         self,
         cache_path: str,
+        log_path: Optional[str],
         verbosity: int,
         colors: Optional[bool],
         n_jobs: int,
@@ -126,6 +133,13 @@ class Config:
         fail_fast: bool,
     ) -> None:
         self.cache_path = Path(cache_path).resolve()
+        if log_path is None:
+            self.log_path = self.cache_path / "logs"
+        else:
+            self.log_path = Path(log_path).resolve()
+
+        self._prepare_and_clean_log_path()
+
         self.venvs_path = self.cache_path / "venvs"
 
         self.verbosity = verbosity
@@ -140,6 +154,10 @@ class Config:
             n_jobs = multiprocessing.cpu_count()
         self.n_jobs = n_jobs
 
+        self.is_interactive = (
+            sys.__stdout__.isatty() and sys.__stderr__.isatty()
+        )
+
         self.environ = {
             # XXX: keep this list in sync with the above documentation
             key: os.environ[key]
@@ -211,4 +229,13 @@ class Config:
         if "GITHUB_ACTION" in os.environ:
             return True
 
-        return sys.stdin.isatty()
+        return self.is_interactive
+
+    def _prepare_and_clean_log_path(self) -> None:
+        self.log_path.mkdir(parents=True, exist_ok=True)
+
+        for file in self.log_path.glob("*"):
+            if file.is_dir():
+                shutil.rmtree(file)
+            else:
+                file.unlink()
