@@ -1,15 +1,11 @@
 import logging
-import re
-import sys
 from contextvars import ContextVar
 from types import TracebackType
-from typing import Any, Optional, Tuple, Type, Union, cast
+from typing import Any, Optional, TextIO, Tuple, Type, Union, cast
 
 from colorama import Back, Fore, Style, init
 
-from ._log_capture import WriterProtocol
-
-_ContextHandler = ContextVar[logging.Handler]("_ContextHandler")
+from ._io import ANSI_ESCAPE_CODE_RE
 
 
 class ColorFormatter(logging.Formatter):
@@ -40,53 +36,52 @@ class ColorFormatter(logging.Formatter):
 
 
 class NoColorFormatter(logging.Formatter):
-    ESCAPE_CODE = re.compile(r"\x1b\[\d+m")
-
     def formatMessage(self, record: logging.LogRecord) -> str:
         msg = super().formatMessage(record)
-        return self.ESCAPE_CODE.sub("", msg)
+        return ANSI_ESCAPE_CODE_RE.sub("", msg)
 
 
-class ContextBasedHandler(logging.Handler):
-    def __init__(self, level: int = logging.NOTSET) -> None:
-        super().__init__(level)
-        self._default = _create_handler(sys.stderr)
+class ContextStreamHandler(logging.StreamHandler):  # type: ignore[type-arg]
+    _stream: ContextVar[TextIO]
 
-    def emit(self, record: logging.LogRecord) -> None:
-        try:
-            handler = _ContextHandler.get()
-        except (
-            LookupError
-        ):  # There is no way of setting a handler after the fact
-            handler = self._default
+    @property  # type: ignore[override]
+    def stream(self) -> TextIO:
+        return self._stream.get()
 
-        handler.setFormatter(self.formatter)
-        handler.emit(record)
+    @stream.setter
+    def stream(self, value: ContextVar[TextIO]) -> None:
+        self._stream = value
 
 
-def _create_handler(output: WriterProtocol) -> logging.Handler:
-    return logging.StreamHandler(output)
+def setup_logging(
+    level: int,
+    colors: bool,
+    tty_output: ContextVar[TextIO],
+    log_file: ContextVar[TextIO],
+) -> None:
+    nocolor_formatter = NoColorFormatter(
+        fmt="dwas > [%(levelname)s] %(message)s"
+    )
 
-
-def setup_logging(level: int, colors: bool) -> None:
     if colors:
         init(strip=False)
-        formatter: logging.Formatter = ColorFormatter(
+        stderr_formatter: logging.Formatter = ColorFormatter(
             fmt=(
                 f"{Fore.CYAN}{Style.DIM}dwas >{Style.RESET_ALL}"
                 f" %(level_color)s%(message)s{Style.RESET_ALL}"
             )
         )
     else:
-        formatter = NoColorFormatter(fmt="dwas > [%(levelname)s] %(message)s")
+        stderr_formatter = nocolor_formatter
 
     logger = logging.getLogger()
-    logger.setLevel(level)
+    logger.setLevel(logging.NOTSET)
 
-    handler = ContextBasedHandler()
-    handler.setFormatter(formatter)
-    logger.addHandler(handler)
+    stderr_handler = ContextStreamHandler(tty_output)
+    stderr_handler.setLevel(level)
+    stderr_handler.setFormatter(stderr_formatter)
+    logger.addHandler(stderr_handler)
 
-
-def set_context_handler(output: WriterProtocol) -> None:
-    _ContextHandler.set(_create_handler(output))
+    logfile_handler = ContextStreamHandler(log_file)
+    logfile_handler.setFormatter(nocolor_formatter)
+    logger.addHandler(logfile_handler)
