@@ -1,11 +1,12 @@
-import graphlib
+from __future__ import annotations
+
 import logging
 import signal
-import sys
 import time
 from collections import deque
 from concurrent import futures
-from contextvars import Context, ContextVar, copy_context
+from contextlib import ExitStack
+from contextvars import ContextVar, copy_context
 from datetime import timedelta
 from subprocess import CalledProcessError
 from types import FrameType
@@ -17,18 +18,18 @@ from dwas._steps.handlers import BaseStepHandler, StepGroupHandler, StepHandler
 from dwas._steps.parametrize import extract_parameters
 from dwas._steps.steps import Step
 
+from . import _io
 from ._config import Config
 from ._exceptions import (
-    CyclicStepDependenciesException,
     DuplicateStepException,
     FailedPipelineException,
     UnavailableInterpreterException,
     UnknownStepsException,
 )
-from ._log_capture import PipePlexer
-from ._logging import set_context_handler
-from ._subproc import ProcessManager, set_subprocess_default_pipes
-from ._timing import get_timedelta_since
+from ._frontend import Frontend, StepSummary
+from ._scheduler import JobResult, Resolver, Scheduler
+from ._subproc import ProcessManager
+from ._timing import format_timedelta, get_timedelta_since
 
 LOGGER = logging.getLogger(__name__)
 
@@ -141,7 +142,7 @@ class Pipeline:
                 description=current_description,
                 func=func,
                 pipeline=self,
-                python=args.pop("python", None),
+                python_spec=args.pop("python", None),
                 requires=args.pop("requires", None),
                 run_by_default=current_run_by_default,
                 parameters=args,
@@ -282,16 +283,6 @@ class Pipeline:
 
         return graph
 
-    def _resolve_execution_order(
-        self, graph: Dict[str, List[str]]
-    ) -> List[str]:
-        sorter = graphlib.TopologicalSorter(graph)
-
-        try:
-            return list(sorter.static_order())
-        except graphlib.CycleError as exc:
-            raise CyclicStepDependenciesException(exc.args[1]) from exc
-
     def execute(
         self,
         steps: Optional[List[str]],
@@ -299,12 +290,13 @@ class Pipeline:
         only_selected_steps: bool,
         clean: bool,
     ) -> None:
-        # we should refactor at some point
-        # pylint: disable=too-many-branches,too-many-locals,too-many-statements
+        # pylint: disable=too-many-locals
         start_time = time.monotonic()
 
         graph = self._build_graph(steps, except_steps, only_selected_steps)
-        steps_in_order = self._resolve_execution_order(graph)
+        resolver = Resolver(graph)
+        scheduler = resolver.get_scheduler()
+        steps_in_order = resolver.order()
 
         if clean:
             LOGGER.debug("Cleaning up workspace")
@@ -313,29 +305,18 @@ class Pipeline:
 
         LOGGER.info("Running steps: %s", ", ".join(steps_in_order))
 
-        sorter = graphlib.TopologicalSorter(graph)
-        sorter.prepare()
-
         should_stop = False
-        running_futures: Dict[
-            futures.Future[timedelta], Tuple[str, Optional[PipePlexer]]
-        ] = {}
-
-        def stop() -> None:
-            nonlocal should_stop
-            should_stop = True
-            for future, (name, _) in running_futures.items():
-                LOGGER.debug("Cancelling step %s", name)
-                future.cancel()
 
         def request_stop(_signum: int, _frame: Optional[FrameType]) -> None:
+            nonlocal should_stop
+
             if not should_stop:
                 LOGGER.warning(
                     "%sStopping requested. This will finish current jobs."
                     " To abort, hit ^C a second time.",
                     Style.BRIGHT,
                 )
-                stop()
+                should_stop = True
             else:
                 LOGGER.critical("Aborting")
                 self.proc_manager.kill()
@@ -343,13 +324,17 @@ class Pipeline:
         previous_signal = signal.signal(signal.SIGINT, request_stop)
 
         try:
-            results = self._execute(
-                sorter, running_futures, stop, lambda: should_stop
-            )
+            with ExitStack() as stack:
+                if self.config.is_interactive:
+                    stack.enter_context(
+                        Frontend(StepSummary(scheduler, start_time)).activate()
+                    )
+
+                self._execute(scheduler, lambda: should_stop)
         finally:
             signal.signal(signal.SIGINT, previous_signal)
 
-        self._log_summary(graph, results, start_time)
+        self._log_summary(scheduler, steps_in_order, graph, start_time)
 
     def get_step(self, step_name: str) -> BaseStepHandler:
         return self.steps[step_name]
@@ -362,12 +347,12 @@ class Pipeline:
         show_dependencies: bool = False,
     ) -> None:
         # pylint: disable=too-many-locals
-        all_steps = self._resolve_execution_order(
+        all_steps = Resolver(
             self._build_graph(list(self.steps.keys()))
-        )
-        selected_steps = self._resolve_execution_order(
+        ).order()
+        selected_steps = Resolver(
             self._build_graph(steps, except_steps, only_selected_steps)
-        )
+        ).order()
 
         step_infos = []
         for step in sorted(all_steps):
@@ -417,157 +402,170 @@ class Pipeline:
 
     def _execute(
         self,
-        sorter: "graphlib.TopologicalSorter[str]",
-        running_futures: Dict[
-            futures.Future[timedelta], Tuple[str, Optional[PipePlexer]]
-        ],
-        stop: Callable[[], None],
+        scheduler: Scheduler,
         should_stop: Callable[[], bool],
-    ) -> Dict[str, Tuple[Optional[Exception], timedelta]]:
-        results: Dict[str, Tuple[Optional[Exception], timedelta]] = {}
+    ) -> None:
+        running_futures: Dict[
+            futures.Future[None], Tuple[str, Optional[_io.PipePlexer]]
+        ] = {}
 
         with futures.ThreadPoolExecutor(self.config.n_jobs) as executor:
-            while sorter.is_active():
-                ready = sorter.get_ready()
-                if not ready:
-                    if not running_futures:
-                        # No more tasks ready to run, and all tasks has finished
-                        # we can't move forward
-                        break
-
-                    next_finished = next(
-                        futures.as_completed(running_futures.keys())
-                    )
-                    name, pipe_plexer = running_futures.pop(next_finished)
-
-                    if pipe_plexer is not None:
-                        pipe_plexer.dump(sys.stdout, sys.stderr)
-
-                    try:
-                        time_spent = next_finished.result()
-                    except ExceptionWithTimeSpentException as exc:
-                        results[name] = exc.original_exception, exc.time_spent
-
-                        if (
-                            isinstance(
-                                exc.original_exception,
-                                UnavailableInterpreterException,
-                            )
-                            and self.config.skip_missing_interpreters
-                        ):
-                            sorter.done(name)
-                        else:
-                            if self.config.fail_fast:
-                                stop()
-
-                            # We won't be able to enqueue new results after a failure
-                            # anyways
-                            continue
-                    except futures.CancelledError as exc:
-                        results[name] = exc, timedelta()
-                        continue
-                    else:
-                        results[name] = None, time_spent
-                        sorter.done(name)
-
+            while scheduler:
+                # Check if we should stop, and cancel new jobs
                 if should_stop():
+                    scheduler.stop()
+
+                # Enqueue new jobs if possible
+                while (
+                    len(running_futures) < self.config.n_jobs
+                    and scheduler.ready
+                ):
+                    step = scheduler.ready[0]
+                    scheduler.mark_started(step)
+
+                    pipe_plexer = (
+                        _io.PipePlexer() if self.config.n_jobs != 1 else None
+                    )
+
+                    future = cast(
+                        "futures.Future[None]",
+                        executor.submit(
+                            # XXX: mypy gets confused here, but the result is
+                            #      sane
+                            copy_context().run,  # type: ignore[arg-type]
+                            self._run_step,  # type: ignore[arg-type]
+                            step,  # type: ignore[arg-type]
+                            pipe_plexer,  # type: ignore[arg-type]
+                        ),
+                    )
+                    running_futures[future] = step, pipe_plexer
+
+                # Wait for previous jobs to finish
+                if not running_futures:
                     continue
 
-                for name in ready:
-                    pipe_plexer = (
-                        PipePlexer() if self.config.n_jobs != 1 else None
+                next_finished = next(
+                    futures.as_completed(running_futures.keys())
+                )
+                name, pipe_plexer = running_futures.pop(next_finished)
+
+                if pipe_plexer is not None:
+                    with _io.log_file(None):
+                        pipe_plexer.flush()
+
+                self._handle_step_result(name, next_finished, scheduler)
+
+    def _handle_step_result(
+        self,
+        name: str,
+        result: futures.Future[None],
+        scheduler: Scheduler,
+    ) -> None:
+        try:
+            result.result()
+        except Exception as exc:  # pylint: disable=broad-exception-caught
+            if (
+                isinstance(exc, UnavailableInterpreterException)
+                and self.config.skip_missing_interpreters
+            ):
+                scheduler.mark_skipped(name, exc)
+                LOGGER.warning(
+                    "Step %s%s%s failed: %s",
+                    Style.BRIGHT,
+                    name,
+                    Style.NORMAL,
+                    exc,
+                )
+            else:
+                scheduler.mark_failed(name, exc)
+                # FIXME: allow another exception that can be thrown programatically
+                exc_info = (
+                    exc
+                    if not isinstance(
+                        exc,
+                        (CalledProcessError, UnavailableInterpreterException),
                     )
+                    else None
+                )
+                LOGGER.error(
+                    "Step %s failed: %s",
+                    name,
+                    self._format_exception(exc),
+                    exc_info=exc_info,
+                )
 
-                    # XXX: Save the context to be able to rerun in it with each
-                    # executor. This needs to be done every time as it's not
-                    # possible to re-enter a context.
-                    pipeline_context = copy_context()
-
-                    future = executor.submit(
-                        self._run_step_in_context,
-                        pipeline_context,
-                        name,
-                        pipe_plexer,
-                    )
-                    running_futures[future] = name, pipe_plexer
-
-        return results
+                if self.config.fail_fast:
+                    scheduler.stop()
+        else:
+            scheduler.mark_success(name)
+            LOGGER.info(
+                "%sStep %s%s%s finished successfully",
+                Fore.GREEN,
+                Style.BRIGHT,
+                name,
+                Style.NORMAL,
+            )
 
     def _log_summary(
         self,
+        scheduler: Scheduler,
+        steps_order: List[str],
         graph: Dict[str, List[str]],
-        results: Dict[str, Tuple[Optional[Exception], timedelta]],
         start_time: float,
     ) -> None:
         LOGGER.info("%s*** Steps summary ***", Style.BRIGHT)
-        sorter = graphlib.TopologicalSorter(graph)
 
-        failed_jobs = []
-        blocked_jobs = []
-        cancelled_jobs = []
+        for name in steps_order:
+            result, time_spent, exc = scheduler.results[name]
+
+            if result == JobResult.SUCCESS:
+                assert time_spent is not None
 
-        for name in sorter.static_order():
-            if name in results:
-                result, time_spent = results[name]
-                if result is None:
-                    LOGGER.info(
-                        "\t%s[%s] %s%s%s: success",
-                        Fore.GREEN,
-                        time_spent,
-                        Style.BRIGHT,
-                        name,
-                        Style.NORMAL,
-                    )
-                elif (
-                    isinstance(result, UnavailableInterpreterException)
-                    and self.config.skip_missing_interpreters
-                ):
-                    LOGGER.info(
-                        "\t%s[-:--:--.------] %s%s%s: Skipped: %s",
-                        Fore.YELLOW,
-                        Style.BRIGHT,
-                        name,
-                        Style.NORMAL,
-                        result,
-                    )
-                elif isinstance(result, futures.CancelledError):
-                    LOGGER.info(
-                        "\t%s[-:--:--.------] %s%s%s: Cancelled",
-                        Fore.YELLOW,
-                        Style.BRIGHT,
-                        name,
-                        Style.NORMAL,
-                    )
-                    cancelled_jobs.append(name)
-                else:
-                    LOGGER.info(
-                        "\t%s%s[%s] %s: error: %s",
-                        Style.BRIGHT,
-                        Fore.RED,
-                        time_spent,
-                        name,
-                        self._format_exception(result),
-                    )
-                    failed_jobs.append(name)
-            elif cancelled_jobs:
                 LOGGER.info(
-                    "\t%s[-:--:--.------] %s%s%s: Cancelled",
+                    "\t%s[%s] %s%s%s: success",
+                    Fore.GREEN,
+                    format_timedelta(time_spent),
+                    Style.BRIGHT,
+                    name,
+                    Style.NORMAL,
+                )
+            elif result == JobResult.SKIPPED:
+                LOGGER.info(
+                    "\t%s[-:--:--] %s%s%s: Skipped: %s",
+                    Fore.YELLOW,
+                    Style.BRIGHT,
+                    name,
+                    Style.NORMAL,
+                    exc,
+                )
+            elif result == JobResult.CANCELLED:
+                LOGGER.info(
+                    "\t%s[-:--:--] %s%s%s: Cancelled",
                     Fore.YELLOW,
                     Style.BRIGHT,
                     name,
                     Style.NORMAL,
                 )
-                cancelled_jobs.append(name)
-            else:
+            elif result == JobResult.FAILED:
+                assert time_spent is not None
+                assert exc is not None
+
+                LOGGER.info(
+                    "\t%s%s[%s] %s: error: %s",
+                    Style.BRIGHT,
+                    Fore.RED,
+                    format_timedelta(time_spent),
+                    name,
+                    self._format_exception(exc),
+                )
+            elif result == JobResult.BLOCKED:
                 blocking_dependencies = [
                     dep
                     for dep in graph[name]
-                    if dep in failed_jobs or dep in blocked_jobs
+                    if dep in scheduler.failed or dep in scheduler.blocked
                 ]
-                assert blocking_dependencies is not None
-                blocked_jobs.append(name)
                 LOGGER.warning(
-                    "\t%s[-:--:--.------] %s%s%s: blocked by %s",
+                    "\t%s[-:--:--] %s%s%s: blocked by %s",
                     Fore.YELLOW,
                     Style.BRIGHT,
                     name,
@@ -575,78 +573,41 @@ class Pipeline:
                     ", ".join(blocking_dependencies),
                 )
 
-        self._display_slowest_dependency_chain(graph, results)
+        self._display_slowest_dependency_chain(graph, scheduler.results)
 
-        LOGGER.info("All steps ran in %s", get_timedelta_since(start_time))
-        if failed_jobs:
+        LOGGER.info(
+            "All steps ran in %s",
+            format_timedelta(get_timedelta_since(start_time)),
+        )
+        if scheduler.failed or scheduler.blocked or scheduler.cancelled:
             raise FailedPipelineException(
-                len(failed_jobs), len(blocked_jobs), len(cancelled_jobs)
+                len(scheduler.failed),
+                len(scheduler.blocked),
+                len(scheduler.cancelled),
             )
 
-    def _run_step_in_context(
-        self,
-        pipeline_context: Context,
-        name: str,
-        pipe_plexer: Optional[PipePlexer],
-    ) -> timedelta:
-        # We need to make sure that we run in a sub-context of the pipeline.
-        # Running via `futures.ThreadPoolExecutor` however does not forward the
-        # context. As such, we need first to re-enter the pipeline context,
-        # copy it, to avoid modifications, and run inside this new context.
-        def execute() -> timedelta:
-            # Ensure we run in a clean sub-context
-            context = copy_context()
-            return context.run(self._run_step, name, pipe_plexer)
-
-        # Forcefully run in the pipeline context
-        return pipeline_context.run(execute)
-
     def _run_step(
         self,
         name: str,
-        pipe_plexer: Optional[PipePlexer],
-    ) -> timedelta:
-        if pipe_plexer is not None:
-            set_context_handler(pipe_plexer.stderr)
-            set_subprocess_default_pipes(
-                pipe_plexer.stdout, pipe_plexer.stderr
+        pipe_plexer: Optional[_io.PipePlexer],
+    ) -> None:
+        with ExitStack() as stack:
+            if pipe_plexer is not None:
+                stack.enter_context(
+                    _io.redirect_streams(
+                        pipe_plexer.stdout, pipe_plexer.stderr
+                    )
+                )
+
+            LOGGER.info(
+                "%s--- Step: %s ---%s", Style.BRIGHT, name, Style.RESET_ALL
             )
+            log_file = self.config.log_path / f"{name}.log"
 
-        LOGGER.info(
-            "%s--- Step: %s ---%s", Style.BRIGHT, name, Style.RESET_ALL
-        )
-        start_time = time.monotonic()
+            LOGGER.debug("Log file can be found at %s", log_file)
+            stack.enter_context(_io.log_file(log_file))
 
-        try:
             self.steps[name].execute()
-        except UnavailableInterpreterException as exc:
-            LOGGER.warning(
-                "Step %s%s%s failed: %s", Style.BRIGHT, name, Style.NORMAL, exc
-            )
-            raise ExceptionWithTimeSpentException(
-                exc, get_timedelta_since(start_time)
-            ) from exc
-        except Exception as exc:
-            # FIXME: allow another exception that can be thrown programatically
-            exc_info = exc if not isinstance(exc, CalledProcessError) else None
-            LOGGER.error(
-                "Step %s failed: %s",
-                name,
-                self._format_exception(exc),
-                exc_info=exc_info,
-            )
-            raise ExceptionWithTimeSpentException(
-                exc, get_timedelta_since(start_time)
-            ) from exc
-
-        LOGGER.info(
-            "%sStep %s%s%s finished successfully",
-            Fore.GREEN,
-            Style.BRIGHT,
-            name,
-            Style.NORMAL,
-        )
-        return get_timedelta_since(start_time)
 
     def _format_exception(self, exc: Exception) -> str:
         if isinstance(exc, CalledProcessError):
@@ -656,7 +617,7 @@ class Pipeline:
     def _display_slowest_dependency_chain(
         self,
         graph: Dict[str, List[str]],
-        results: Dict[str, Tuple[Optional[Exception], timedelta]],
+        results: Dict[str, Tuple[JobResult, timedelta, Optional[Exception]]],
     ) -> None:
         if len(graph) <= 1:
             # If there's a single entry in the whole graph, no need to show
@@ -691,7 +652,7 @@ class Pipeline:
     def _compute_slowest_chains(
         self,
         graph: Dict[str, List[str]],
-        results: Dict[str, Tuple[Optional[Exception], timedelta]],
+        results: Dict[str, Tuple[JobResult, timedelta, Optional[Exception]]],
     ) -> Dict[str, Tuple[List[str], timedelta]]:
         total_time_per_step: Dict[str, Tuple[List[str], timedelta]] = {}
 
